"""
This file is part of Lynq (elemenom/lynq).

Lynq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lynq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Lynq. If not, see <https://www.gnu.org/licenses/>.
"""

import atexit, os, argparse, logging, json

from lynq.backendutils.lynq.pycache_remover import remove_pycache_from as _remove_pycache_from

from lynq.backendutils.basin.getval import getval
from lynq.backendutils.basin.object import BasinObject

from typing import Any

from sysver import VERSION, INSTALL, UPGRADE

warn, warn2 = False, False

_, _ = warn, warn2 # Bypass 'Redeclared unused variable' warning.

parser: argparse.ArgumentParser = argparse.ArgumentParser(description="Basic config")

parser.add_argument("--lq.cfile", type=str, help="Path to your lynqconfig file.")
parser.add_argument("--lq.ctype", type=str, help="Type of your lynqconfig file. Supports 'JSON', 'BASIN' and 'PYTHON'.")

parser.add_argument("--lq.rungui", action="store_true", help="Start a new instance of Lynq RUNGUI and exit the program.")
parser.add_argument("--lq.clean", action="store_true", help="Clean pycache files and exit the program.")

parser.add_argument("--lq.run_process", type=str, help="Run a process like from a Lynq RUNGUI, but presented directly from a CLI.")

args: argparse.Namespace = parser.parse_args()

match args.__getattribute__("lq.ctype"):
    case "JSON":
        with open(args.__getattribute__("lq.cfile")) as file:
            data: dict[str, Any] = json.load(file)

        logger = eval(data.get("logger", "None"))
        additional = eval(data.get("loggingConfig", "None"))
        level = data.get("loggingLevel", None)
        format_ = data.get("loggingFormat", None)

        cleanlogger = data.get("cleanLogger", "None")
        clean = data.get("cleanPycache", "None")
        cleanlogfile = data.get("cleanLogFile", "None")

    case "BASIN":
        data: BasinObject = BasinObject(args.__getattribute__("lq.cfile"))

        logger = eval(getval("logger", data))
        additional = eval(getval("loggingConfig", data))
        level = getval("loggingLevel", data)
        format_ = getval("loggingFormat", data)

        cleanlogger = eval(getval("cleanLogger", data, "None").title())
        clean = eval(getval("cleanPycache", data, "None").title())
        cleanlogfile = eval(getval("cleanLogFile", data, "None").title())

    case "PYTHON":
        try:
            from lynqconfig import logger as logger # type: ignore
            from lynqconfig import loggingConfig as additional # type: ignore
            from lynqconfig import loggingLevel as level # type: ignore
            from lynqconfig import loggingFormat as format_ # type: ignore

            from lynqconfig import cleanLogger as cleanlogger # type: ignore
            from lynqconfig import cleanPycache as clean # type: ignore
            from lynqconfig import cleanLogFile as cleanlogfile # type: ignore
        except (ModuleNotFoundError, ImportError):
            logger, \
            additional, \
            level, \
            format_, \
            clean, \
            cleanlogger, \
            cleanlogfile \
            = (None,) * 7

            warn = True

    case _:
        logger, \
        additional, \
        level, \
        format_, \
        clean, \
        cleanlogger, \
        cleanlogfile \
        = None, None, None, None, None, None, None

        warn2 = True

PYCACHE_REMOVAL_LOCATIONS: tuple[str, ...] = (
    "",
    "backendutils",
    "backendutils.app",
    "backendutils.basin",
    "backendutils.custom",
    "backendutils.dependencies",
    "backendutils.launcher",
    "backendutils.lynq",
    "backendutils.server",
    "backendutils.style"
)

class MyLynq:
    class get_version:
        @staticmethod
        def whole() -> float:
            return VERSION["whole"]
        
        @staticmethod
        def major() -> int:
            return VERSION["major"]
        
        @staticmethod
        def minor() -> float:
            return VERSION["minor"]
        
    class install:
        @staticmethod
        def pip() -> str:
            return INSTALL["pip"]
        
        @staticmethod
        def git() -> str:
            return INSTALL["git"]
        
    class upgrade:
        @staticmethod
        def pip() -> str:
            return UPGRADE["pip"]
        
        @staticmethod
        def git() -> str:
            return UPGRADE["git"]
        
    @staticmethod
    def help() -> str:
        with open("README.md") as file_:
            return file_.read()
        
    @staticmethod
    def send_feedback() -> None:
        print("Have some feedback regarding Lynq? Great! Submit an issue at https://github.com/elemenom/lynq/issues/new")

    @staticmethod
    def license() -> str:
        with open("LICENSE") as file_:
            return file_.read()

    @staticmethod
    def get_sysver() -> float:
        return MyLynq.get_version.whole()

logging.basicConfig(
    level = eval(f"logging.{level}") if level else logging.DEBUG,
    format = format_ or "%(asctime)s ~ %(levelname)s | %(message)s",
    **additional or {}
)

GLOBAL_LOGGER: Any = logger or logging.getLogger(__name__)
CLEAN_CACHE: bool = clean or False
CLEAN_LOGGER: bool = cleanlogger or True

GLOBAL_LOGGER.info(f"Started instance of Lynq v{VERSION["major"]}(.{int(VERSION["minor"] * 10)})")

from lynq.backendutils.launcher.launch import launch
from lynq.backendutils.launcher.direct import directlaunch
from lynq.backendutils.app.app import app
from lynq.backendutils.app.jsonapp import jsonapp
from lynq.backendutils.app.basinapp import basinapp
from lynq.backendutils.lynq.msie import InternetExplorerInstance
from lynq.backendutils.server.standard import LynqServer
from lynq.backendutils.server.custom import ConfigurableLynqServer
from lynq.backendutils.server.json import JsonLynqServer
from lynq.backendutils.server.basin import BasinLynqServer
from lynq.backendutils.app.appobject import AppObject as App
from lynq.backendutils.app.standardappexportobject import StandardAppExportObject as ExportedApp
from lynq.backendutils.script.ctrl import CTRLScript

if args.__getattribute__("lq.rungui"):
    from lynq.backendutils.lynq.rungui import run_gui

    run_gui("<onstart rungui>")

    exit()

if args.__getattribute__("lq.clean"):
    for path in PYCACHE_REMOVAL_LOCATIONS:
        _remove_pycache_from(f"./lynq/{path.replace(".", "/")}")

    exit()

if args.__getattribute__("lq.run_process"):
    from lynq.backendutils.lynq.rungui import run_process

    run_process(args.__getattribute__("lq.run_process"), GLOBAL_LOGGER, "<terminal (ran using lynq CLI)>")

    exit()

if warn:
    GLOBAL_LOGGER.error("An error occured in your lynqconfig PYTHON file. All config will be ignored (default will be used for everything). Make sure you include ALL configurements, and set them to `None` if you don't need to change them.")
    GLOBAL_LOGGER.error("PLEASE NOTE THAT IF THE CONFIG TYPE IS 'PYTHON', WE ALWAYS USE THE './lynqconfig.py' PATH; THE 'configfile' ARGUMENT IS IGNORED. * This does not apply to JSON and BASIN type lynqconfig files.")

if warn2:
    GLOBAL_LOGGER.error("None or invalid lynqconfig type provided in args, returned all options to their default values.")

def _clean_up() -> None:
    logging.shutdown()

    if os.path.exists("throwaway.log") and cleanlogfile:
        os.remove("throwaway.log")

def _clean_up_cache() -> None:
    GLOBAL_LOGGER.debug("Commencing pycache clean up process.")

    for path_ in PYCACHE_REMOVAL_LOCATIONS:
        _remove_pycache_from(f"./lynq/{path_.replace(".", "/")}")

def _at_exit_func() -> None:

    GLOBAL_LOGGER.debug("Commencing logger deletion and clean up process.")
    
    if CLEAN_CACHE:
        _clean_up_cache()

    if CLEAN_LOGGER:
        _clean_up()

    print(f"[Exiting...] Program ended successfully. All active servers terminated.")

atexit.register(_at_exit_func)


"""
This file is part of Lynq (elemenom/lynq).

Lynq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lynq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Lynq. If not, see <https://www.gnu.org/licenses/>.
"""

import os

from typing import Optional
from typing import Any

from lynq import GLOBAL_LOGGER as logger

from lynq.backendutils.launcher.launch import launch

from lynq.backendutils.app.supportswith import SupportsWithKeyword
from lynq.backendutils.lynq.lynqserverorrelated import LynqServerOrRelatedObjects
from lynq.backendutils.app.supportedtags import supported_tags
from lynq.backendutils.script.ctrl import CTRLScript

from lynq.backendutils.style.style import StyledAppAttachment

class AppObject(SupportsWithKeyword):
    def __init__(self, name: Optional[str], app: Any = None) -> None:
        from lynq.backendutils.app.app import app as app_

        self.server: Any = app.server
        self.name: str = f"{name}.html"
        self.app: app_ | None = app
        self.style_: StyledAppAttachment | None = app.style
        self.new_stylesheet_path: str = f"{self.app.fn.__name__}.css"
        self.ctrl: CTRLScript | None = None

        self.init_supported_tags(supported_tags)
        self.init_ctrlscript()

        self.on_run()

    def on_run(self) -> None:
        logger.info("Please wait while we dist your HTML file for you.")

        self.singular("<!DOCTYPE html>")
        self.singular("<html>")
        self.singular(f"<link rel=\"stylesheet\" href=\"{self.new_stylesheet_path}\">")
        self.singular(f"<script src=\"{self.app.fn.__name__}.js\"></script>")

    def init_ctrlscript(self) -> None:
        self.ctrl = CTRLScript(self.app.fn.__name__)

    def init_supported_tags(self, supported_tags_: list[str]) -> None:
        for tag in supported_tags_:
            exec(f"self.{tag} = lambda **args: self.tag({repr(tag)}, **args)", {"self": self})

    def tag(self, type_: str, **kwargs: str | int) -> Any:
        from lynq.backendutils.app.tagobject import TagObject

        final_kwargs = ""

        for kwarg in list(kwargs.items()):
            final_kwargs += f"{kwarg[0]}=\"{kwarg[1]}\" "

        return TagObject(self.name.removesuffix(".html"), type_, self.app, final_kwargs)
    
    def singular(self, ln: str) -> None:
        with open(self.name, "a") as file:
            file.write(ln + "\n")
    
    def __exit__(self, *_) -> None:
        self.exit()

    def exit(self) -> None:
        self.singular("</html>")

        logger.info("Building has been finished successfully.")
    
    def pass_to_server(self) -> None:
        if self.server is None:
            logger.fatal("Cannot pass pebl script to server when no server was provided.")
            exit(1)

        logger.info(f"Passed '{self.name}' pebl script to host '{type(self.server).__name__}'")

        if self.style is not None:
            logger.info(f"Passed stylesheet '{self.new_stylesheet_path}' to pebl script '{self.name}'.")

            self.style_.close()

        launch(self.server)

        logger.info("Continuing in pebl app to clear cache.")

        try: os.remove(self.name)
        except FileNotFoundError:
            logger.warn(f"Could not find '{self.name}', so it cannot be cleaned.")

        try: os.remove(self.ctrl.path)
        except FileNotFoundError:
            logger.warn(f"Could not find '{self.name}', so it cannot be cleaned.")

        if self.style_ is not None:
            try: os.remove(self.new_stylesheet_path)
            except FileNotFoundError:
                logger.warn(f"Could not find '{self.new_stylesheet_path}', so it cannot be cleaned.")


"""
This file is part of Lynq (elemenom/lynq).

Lynq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lynq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Lynq. If not, see <https://www.gnu.org/licenses/>.
"""

from tkinter import Tk, Label, Entry, Button
from tkinter.messagebox import showerror, showinfo

from typing import Any, Optional

from lynq.backendutils.lynq.msie import pwsh

def run_process(text: str, logger: Any, id: Optional[str] = None) -> None:
    if text == "":
        showerror("Error", "Please enter a value. Trying to specify a FilePath in the terminal? Try running '$$TERMINAL'. (without quotes)")

        return

    elif text.upper() == "$$TERMINAL":
        print("/!\\ Continuing in a terminal does not support '$$' commands, '>' commands, '%' commands, 'lynq://' sites, and custom commands like 'lynqping', 'exit' or 'lynqhelp'.")
        text = ""

    elif text.upper() == "$$HELP":
        text = ">helplynq"
        
    elif text.upper() == "$$PING":
        text = ">pinglynq"

    elif text.upper() == "$$EXIT":
        text = ">exit"

    elif text.upper() == "$$INDEX":
        text = "lynq://app/index.py"

    elif text.upper() == "$$-8000":
        text = "lynq://8000"

    elif text.upper() == "$$-24001":
        text = "lynq://24001"

    if text.startswith("lynq://app/"):
        pwsh(f"python {text.removeprefix("lynq://app/")}.py")

    elif text.startswith("lynq://"):
        text = f"http://localhost:{text.removeprefix("lynq://")}/"

        pwsh(f"start {text}")

    elif (text.strip(">") == "exit") or (text == r"%exit"):
        logger.info("Exited Lynq RUNGUI.")

        entry.master.destroy()

    elif (text.strip(">") == "help") or (text == r"%help"):
        print("> Looking for help regarding Lynq and Lynq RUNGUI? Try '$$HELP'.")

    elif (text.strip(">") == "helplynq") or (text == r"%helplynq"):
        showinfo("Help", r"$$EXIT, %exit, >exit - exit the GUI and end the program. $$PING, %pinglynq, >pinglynq - Ping the terminal from this Lynq RUNGUI. $$TERMINAL - continue in terminal. See the 'README.md' file or run 'import lynq; print(lynq.MyLynq.help())' in a Python command line interface for more help.")

    elif (text.strip(">") == "pinglynq") or (text == r"%pinglynq"):
        logger.info(f"Ping received from Lynq RUNGUI 'id={id or "none (unnamed)"}'.")

    elif text.startswith(">") or text.startswith("%"):
        pwsh(text.strip(">") if text.strip(">") != text else text.removeprefix("%"))

    else:
        pwsh(f"start {text}")

def run_gui(id: Optional[str] = None) -> None:
    from lynq import GLOBAL_LOGGER as logger

    root: Tk = Tk()

    root.title("Run using Lynq")
    root.geometry("400x150")
    root.resizable(False, False)

    Label(root, text="Run a new process using Lynq").pack()
    entry: Entry = Entry(root, width=80)
    entry.pack(padx=10, pady=10)
    Button(root, text="Submit", command=lambda: run_process(entry.get(), logger, id)).pack(pady=10)

    logger.info("Completed dist; launch sequence initiated.")

    root.mainloop()


"""
This file is part of Lynq (elemenom/lynq).

Lynq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lynq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Lynq. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Self, Any

from lynq.backendutils.app.supportswith import SupportsWithKeyword

class StyleAttribute(SupportsWithKeyword):
    def __init__(self, style: Any, tag_name: str) -> None:
        self.style: Any = style

        self.name: str = tag_name

        self.written: str = ""

        self._init_attribute()

    def _init_attribute(self) -> None:
        self._write_line(f"{self.name} " + "{")

    def _write_line(self, contents: str) -> None:
        self.written += f"{contents}\n"

    def add_singular(self, contents: str) -> Self:
        self._write_line(contents)

        return self

    def add_option(self, **kwargs) -> Self:
        for key, value in list(kwargs.items()):
            self._write_line(f"{key.replace("_", "-")}: {value};")

        return self

    def __exit__(self, *_) -> None:
        self.close()

    def back(self) -> Self:
        self._write_line("}")

        return self.style

    def get_written(self) -> str:
        return self.written


from typing import Optional

from lynq import GLOBAL_LOGGER as logger

class AppendedFile:
    def __init__(self):
        self.path: str | None = None

    def init_file(self, path: str) -> None:
        self.path = path

        self.write()

    def write(self, cont: Optional[str] = None) -> None:
        if not self.path:
            logger.fatal("Cannot write to file appendant when file was not initialised.")
            exit(1)

        with open(self.path, "a") as file:
            file.write((cont or "") + "\n")

    def read(self) -> str:
        if not self.path:
            logger.fatal("Cannot read from file appendant when file was not initialised.")
            exit(1)

        with open(self.path) as file:
            return file.read()


"""
This file is part of Lynq (elemenom/lynq).

Lynq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lynq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Lynq. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional, Callable

from lynq.backendutils.lynq.lynqserverorrelated import LynqServerOrRelatedObjects
from lynq.backendutils.app.appobject import AppObject
from lynq.backendutils.app.supportswith import SupportsWithKeyword
from lynq.backendutils.app.blankslateobject import new

class app(SupportsWithKeyword):
    def __init__(self, server: Optional[LynqServerOrRelatedObjects] = None, style_path: Optional[str] = None) -> None:
        from lynq.backendutils.style.style import StyledAppAttachment

        self.style = new("StyledAppAttachmentWithAppAttribute", (StyledAppAttachment,), 
            back = lambda _: self
        )(style_path)

        super().__init__()

        self.server: Optional[LynqServerOrRelatedObjects] = server

        self._init_root()

    def _init_root(self) -> None:
        self.export = new("export", (), # Export methods here:
            standard = self.export_standard,
            void = self.export_void,
            void_no_lambda = self.export_voidnolambda,
            direct = self.export_direct,
            dontpass = self.export_nopass
        )

    def export_standard(self, fn: Callable) -> Callable:
        self.fn: Callable = fn

        from lynq.backendutils.app.standardappexportobject import StandardAppExportObject

        return lambda *args, **kwargs: StandardAppExportObject(self, *args, object_=AppObject, **kwargs)
    
    def export_void(self, fn: Callable) -> Callable:
        return lambda *args, **kwargs: fn(*args, **kwargs)

    def export_voidnolambda(self, fn: Callable) -> None:
        return fn

    def export_direct(self, fn: Callable) -> None:
        app_object: AppObject = AppObject(fn.__name__, self)
        app_object.exit()

        try: return fn(app_object)
        finally: app_object.pass_to_server()
    
    def export_nopass(self, fn: Callable) -> None:
        app_object: AppObject = AppObject(fn.__name__, self)
        app_object.exit()

        return fn(app_object)


"""
This file is part of Lynq (elemenom/lynq).

Lynq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lynq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.0

You should have received a copy of the GNU General Public License
along with Lynq. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional

from lynq.backendutils.app.app import app

from lynq.backendutils.server.basin import BasinLynqServer

class basinapp(app):
    def __init__(self, path: Optional[str] = None) -> None:
        super().__init__(BasinLynqServer(path or "index.bsn"))


from contextlib import contextmanager

from lynq.backendutils.script.appendedfile import AppendedFile

class CTRLScript(AppendedFile):
    def __init__(self, name: str):
        super().__init__()

        self.init_file(f"{name}.js")

    @contextmanager
    def function(self, name: str, *args: str) -> None:
        self.write(f"function {name}({", ".join(list(args))}) ""{")

        yield

        self.write("}")

    @contextmanager
    def export_function(self, name: str, *args: str) -> None:
        self.write(f"function {name}({", ".join(list(args))}) ""{")

        yield

        self.write("}")

    def line(self, ln: str) -> None:
        self.write(f"{ln};")

    def import_module(self, name: str) -> None:
        self.write(f"import {name};")

    def let(self, **kwargs: str) -> None:
        for kwarg in list(kwargs.items()):
            self.write(f"let {kwarg[0]} = {kwarg[1]};")

    def set(self, **kwargs: str) -> None:
        for kwarg in list(kwargs.items()):
            self.write(f"{kwarg[0]} = {kwarg[1]};")

    def increment(self, **kwargs: str) -> None:
        for kwarg in list(kwargs.items()):
            self.write(f"{kwarg[0]} += {kwarg[1]};")

    def decrement(self, **kwargs: str) -> None:
        for kwarg in list(kwargs.items()):
            self.write(f"{kwarg[0]} -= {kwarg[1]};")

    def multiply(self, **kwargs: str) -> None:
        for kwarg in list(kwargs.items()):
            self.write(f"{kwarg[0]} *= {kwarg[1]};")

    def divide(self, **kwargs: str) -> None:
        for kwarg in list(kwargs.items()):
            self.write(f"{kwarg[0]} /= {kwarg[1]};")


"""
This file is part of Lynq (elemenom/lynq).

Lynq is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lynq is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Lynq. If not, see <https://www.gnu.org/licenses/>.
"""

from lynq.backendutils.lynq.lynqserverorrelated import LynqServerOrRelatedObjects

def launch(server: LynqServerOrRelatedObjects) -> LynqServerOrRelatedObjects:
    try:
        server.open()
        input("\033[1;93mPress enter to exit your Lynq server...\n\033[0m")

    finally:
        server.close()

        return server